/*
 * 	 Drop.java
 * 
 * 
 *   famCNV: Copy Number Variant association for quantitative traits in families
 *
 *   famCNV is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.

 *   famCNV is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with famCNV.  If not, see <http://www.gnu.org/licenses/>.
 *
 *   For any bugs or problems found, please contact us at
 *   m.falchi@imperial.ac.uk  
 *	 hariklia.eleftherohorinou06@imperial.ac.uk
 * 	 visconti@di.unito.it
 */

import java.util.*;
import java.io.*;
import java.lang.Runtime;
import java.util.concurrent.*;

public class Drop 
{

	private Hashtable<Integer, Family> families;
	private Vector<Trait> traitNames;
	private Vector<Marker> listaSNP;
	private Vector<String> includedSNP;
	private int numCovar;
	private int numSimul;
	private int tests;
	

	public Drop()
	{
		families = new Hashtable<Integer, Family>();
		traitNames = new Vector<Trait>();
		listaSNP = new Vector<Marker>();
		includedSNP = new Vector<String>();
		numCovar = Utilities.INVALID_I;
		numSimul = 1;
		tests = 0;
	}
	
	public int numFamilies()
	{
		return families.size();
	}


	public void setNumSimul(int numSimul) 
	{
		this.numSimul = numSimul;
	}
	
	public int numCovar()
	{
		return numCovar;
	}

	public int tests()
	{
		return tests;
	}


	public int loadFamilies(String nomeFile, String phenoFile, int expectedTraits)  throws Exception
	{	
		int people = 0;
		int tmpFamily = Utilities.INVALID_I;
		Integer tmpFamObj = null;

		//After this call all data are loaded in RAM
		PedParser inputData; //file parser
		inputData = new PedParser(nomeFile, phenoFile);
		
		//parse and load a Person at a time
		while (!inputData.endOfData()) 
		{
			Person person = inputData.readIndividual();
			if (person.getNumTraits() != expectedTraits)
				throw new Exception("ERROR: line " + (people + 1) + " of .ped file does not match the number of traits. Please check your .ped and .trmap file");
			
			people++;
			
			//load family, associating individues to their own family 
			int personFamily = person.famId();
			if (families.get(personFamily) == null)
				families.put(new Integer(personFamily), new Family(personFamily));

			families.get(personFamily).addPerson(person);
		}
		
		//covariates will be read afterwards
		numCovar = 0;
		return people; 
	}
	
	public void familyPreprocess() throws Exception
	{
		//some work to be done over each family
		for (Enumeration<Family> elements = families.elements(); elements.hasMoreElements();)
		{
			Family current = elements.nextElement();
			//current.sanityCheck();
			try
			{
				current.reindex(false);
				current.sort();
				current.initializeKinMatrix();
				current.kinship();
				if (current.hasMZTwin())
					current.mzTwinAdjust();
			}
			catch(Exception e)
			{
				throw new Exception("Error in family " + current.getId() + " -- " + e.getMessage());
			}
		}
	}


	public Vector<Result> analyze() throws Exception
	{	
		if (numCovar == Utilities.INVALID_I)
			throw new Exception("ERROR: Check --cov parameter or covariates file");

		Vector<Input> inputs = new Vector<Input>();	
		for (int t = 0; t < traitNames.size(); t++) 
			for (int lsi = 0; lsi < listaSNP.size(); lsi++) 
				inputs.add(new Input(t, lsi));
		
		return processInputs(inputs);		
	}


	public Vector<Result> processInputs(Vector<Input> inputs) throws InterruptedException, ExecutionException, Exception
	{

		if (Constants.verbose)
			System.out.print("\t");

		int threads = Runtime.getRuntime().availableProcessors();
		threads = Math.min(threads, Constants.threads);
		
		ExecutorService service = Executors.newFixedThreadPool(threads);

		Vector<Future<Result>> futures = new Vector<Future<Result>>();

		final Semaphore semaphore = new Semaphore(1);
		for (final Input input : inputs)
		{
			Callable<Result> callable = new Callable<Result>()
			{
				public Result call() throws Exception
				{
					int t = input.t();
					int lsi = input.lsi();
					
					Marker mrk = listaSNP.get(lsi);
					Trait trt = traitNames.get(t);
					
					if (mrk.getAnalysis(trt, Constants.window) && mrk.isIncluded(includedSNP)) 
					{
						//System.out.println("Trait: " + trt.getName() + "\tSNP: " + mrk.getName());
						//Trait: 1553972_a_at     SNP: rs2838527
						//critical session
						semaphore.acquire();
							tests++;

							if (Constants.verbose && tests % 10 == 0)
								System.out.print(".");
							if (Constants.verbose && tests % 500 == 0)
								System.out.print(" [" + tests + "]\n\t");
						semaphore.release();
						
						Vector<Family> family_to_analize = new Vector<Family>();
						for (Enumeration<Family> elements = families.elements(); elements.hasMoreElements();)
						{
							Family current = elements.nextElement();
							if (current.isAnalyzable(t, lsi))
								family_to_analize.add(current);
						}
					
						VC vc = new VC(numSimul * family_to_analize.size(), numCovar, t, lsi);

						for (int y = 0; y < numSimul; y++) 
							for (Enumeration<Family> elements = family_to_analize.elements(); elements.hasMoreElements();)
								vc.fill(elements.nextElement());
					
						
						Result result = vc.solve();			
						
						//set this info (plus some more)
						result.setInfo(trt.getName(), mrk.getName(), mrk.getChromosome(), mrk.getPosition(), trt.getPositionFrom(), trt.getPositionTo());
						
						return result;			
					}

					return null;
				}
			};

			futures.add(service.submit(callable));
		}

	    service.shutdown();
	
	    Vector<Result> outputs = new Vector<Result>();
	    for (Future<Result> future : futures)
	    {
			Result result = future.get();
			if (result != null)
				outputs.add(result);
	    }
	    return outputs;
	}


	public int readSNPList() throws Exception
	{
		int counter = 0;
		RandomAccessFile file = new RandomAccessFile(Constants.mrk, "r");
	
		String s;
		int INFO_SIZE = 3; //[chr, name, position]
		
		while ((s = file.readLine()) != null) 
		{
			counter++;
			StringTokenizer t = new StringTokenizer(s, "\t");
			Vector<String> info = new Vector<String>();
			while(t.hasMoreTokens())
				info.add(t.nextToken());
			
			if (info.size() != INFO_SIZE)
				throw new Exception("ERROR: line " + counter + " does not describe a valid marker. Please check your .mrkmap file");
			else 
				listaSNP.add(new Marker(info.get(0), info.get(1), Integer.parseInt(info.get(2))));
		}
		
		return listaSNP.size();
	}
	
	public int readIncludedSNPs() throws Exception
	{
		int counter = 0;
		RandomAccessFile file = new RandomAccessFile(Constants.include, "r");
		String s;
	
		while ((s = file.readLine()) != null) 
		{
			counter++;
			StringTokenizer t = new StringTokenizer(s, "\t");
			Vector<String> info = new Vector<String>();
		
			if (!t.hasMoreTokens())
				throw new Exception("ERROR: line " + counter + " does not describe a valid marker. Please check your .include file");
			else 
				includedSNP.add( t.nextToken() );
		}
		
		return includedSNP.size();
	}
	

	// TRANSCRIPT NAME IS THE SECOND COLUMN OF mydata.trmap FILE
	public int readTraits() throws Exception
	{
	    RandomAccessFile file = new RandomAccessFile(Constants.trt, "r");
		String s;
		int INFO_SIZE = 4; //[name, chr, positionFrom, positionTo]
	
		int counter = 0;
		while ((s = file.readLine()) != null) 
		{
			counter++;
			StringTokenizer t = new StringTokenizer(s, "\t");	
			Vector<String> info = new Vector<String>();
			
			while(t.hasMoreTokens() && info.size() < INFO_SIZE)
				info.add(t.nextToken());
			
			if (info.size() != INFO_SIZE)
				throw new Exception("ERROR: line " + counter + " does not describe a valid trait. Please check your .trmap file");
			else 
				traitNames.add(new Trait(info.get(0), info.get(1), Integer.parseInt(info.get(2)), Integer.parseInt(info.get(3))));
		
		}
		
		return traitNames.size();	
	}

	public void readCovariates() throws Exception
	{
		int counter = 0;
		RandomAccessFile covfile = new RandomAccessFile(Constants.cov, "r");
		int INFO_SIZE = 2;
		String s;
		
		Vector<Double> covariates = new Vector<Double>();
		while ((s = covfile.readLine()) != null) 
		{
			counter++;
			StringTokenizer t = new StringTokenizer(s);
		
			int i = 0;
			int[] info = new int[INFO_SIZE];
			
			while(t.hasMoreTokens() && i < INFO_SIZE) 
			{
				info[i] = Integer.parseInt(t.nextToken());
				i++;
			}
			
			if (i != INFO_SIZE) //fam and individual id are mandatory
				throw new Exception("ERROR: line " + counter + " does not describe a valid covariate. Please check your covariate file");
			
			//If the individual is not present in the dataset (PED file) 
			//the covariate information will be discharged 
			Family fam = families.get(info[0]);
			if (fam == null)
				continue;
				
			Person pers = fam.getIndividualById(info[1]);
			if (pers == null)
				continue;
			
			if (!t.hasMoreTokens()) //no cov info
				throw new Exception("ERROR: line " + counter + " does not describe a valid covariate. Please check your covariate file");

			covariates = new Vector<Double>();
			while (t.hasMoreTokens()) 
			{
				String covariate = new String(t.nextToken());
				// if covariate is missing set affection to missing
				if (covariate.equals("X") || covariate.equals("NA") || covariate.equals("x")) 
					covariates.add(Utilities.INVALID_D);
				else 
					covariates.add(new Double(covariate));
			}
			
			pers.setCovariates(covariates);
			numCovar = covariates.size();
		}
		
		covfile.close();
	}


	public void readSNPs(int expectedSNP) throws Exception
	{
		//With BAF the number of token read is double
		if (Constants.Ball)
			expectedSNP *= 2;
		
		int counter =0;
		RandomAccessFile intfile = new RandomAccessFile(Constants.ints, "r");
		int INFO_SIZE = 2;
		String s;
		
		while ((s = intfile.readLine()) != null) 
		{
			counter++;
			Vector<Double> intensities = new Vector<Double>();
			StringTokenizer t = new StringTokenizer(s);
		
			int i = 0;
			int[] info = new int[INFO_SIZE];
			
			while(t.hasMoreTokens() && i < INFO_SIZE) 
			{
				info[i] = Integer.parseInt(t.nextToken());
				i++;
			}
			
			if (i != INFO_SIZE) //fam and individual id are mandatory
				throw new Exception("ERROR: line " + counter + " does not describe a valid intensity. Please check your .int file");
			
			//If the individual is not present in the dataset (PED file) 
			//the covariate information will be discharged 
			Family fam = families.get(info[0]);
			if (fam == null)
				continue;
				
			Person pers = fam.getIndividualById(info[1]);
			if (pers == null)
				continue;
			
			//we have at least one info, so she is a real person
			pers.setMock(false);
					
			if (!t.hasMoreTokens()) //no intensity info
				throw new Exception("ERROR: line " + counter + " does not describe a valid intensity. Please check your .int file");

			Vector<Double> tmps = new Vector<Double>();
			while (t.hasMoreTokens()) 
			{
				String tmp = new String(t.nextToken());				
				// if covariate is missing set affection to missing
				if (tmp.equals("X") || tmp.equals("NA") || tmp.equals("x")) 
					tmps.add(Utilities.INVALID_D);
				else 
					tmps.add(new Double(tmp));
			}
			
			if (tmps.size() < expectedSNP)
				throw new Exception("ERROR: line " + counter + " of .int does not match the number of markers. Please check your .int and .mrkmap file");
					
			// if Ball == true each CNV has two columns: Ball and LogR, 
			if (Constants.Ball) 
			{
				for (int j = 0; j < tmps.size(); j += 2)
				{
					pers.addBAllele(tmps.get(j));
					pers.addLogR(tmps.get(j+1));
				}	
			} 
			else 
			{	
				for (int j = 0; j < tmps.size(); j++)
				{
					pers.addBAllele(Utilities.INVALID_D);
					pers.addLogR(tmps.get(j));
				}
			}
						
		}
		
		intfile.close();
	}



	public void inverseNormalTransform() throws Exception
	{
		Vector<Person> pedigree = extractPedigree();
		QuantitativeData quantitativeData = new QuantitativeData(pedigree, traitNames.size()); 
		quantitativeData.inverseNormal();
	}


	private Vector<Person> extractPedigree()
	{
		Vector<Person> persons = new Vector<Person>();
		for (Enumeration<Family> elements = families.elements(); elements.hasMoreElements();)
		{
			Family current = elements.nextElement();
			persons.addAll(current.getPersons());
		}

		return persons;
	}

}

class Input
{
	int t;
	int lsi;

	public Input(int t, int lsi)
	{
		this.t = t;
		this.lsi = lsi;
	}

	public int t()
	{
		return t;
	}

	public int lsi()
	{
		return lsi;
	}

}
