/*
 * 	 Family.java
 * 
 * 
 *   famCNV: Copy Number Variant association for quantitative traits in families
 *   Copyright (C) 2010 Dr. Mario Falchi                                   

 *
 *   famCNV is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.

 *   famCNV is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with famCNV.  If not, see <http://www.gnu.org/licenses/>.
 *
 *   For any bugs or problems found, please contact us at
 *   m.falchi@imperial.ac.uk  
 *	 hariklia.eleftherohorinou06@imperial.ac.uk
 *   visconti@di.unito.it
 */

import java.util.*;

class Family 
{

	private Vector<Person> persons;
	private double[][] kinM;
	
	private Vector<Integer> personsId;
	private Vector<Integer> originalIndex;
	private int id;
	private boolean hasMZTwin;

	
	public Family(int id) 
	{
		kinM = null;
		persons = new Vector<Person>();
		personsId = new Vector<Integer>();
		originalIndex = new Vector<Integer>();
		this.id = id;
		hasMZTwin = false;
	}
	
	public int getId()
	{
		return id;
	}
	
	public Vector<Person> getPersons()
	{
		return persons;
	}
	

	public double[][] getKinM()
	{
		return kinM;
	}
	
	
	public Person getIndividual(int numPerson) 
	{
		return persons.get(numPerson);
	}

	public Person getIndividualById(int IDPerson) 
	{
		int index = personsId.indexOf(new Integer(IDPerson));
		if (index == -1)
			return null;
		else
			return persons.elementAt(index);
	}


	public int getNumIndividuals() 
	{
		return persons.size();
	}
	
	public boolean hasMZTwin() 
	{
		return hasMZTwin;
	}


	// aggiunge inizialmente tutte le persone
	// carica come oggetto person e scrive l'id passato come parametro nel
	// vettore personsId
	// la persona viene identificata come progressivo di ingresso, che
	// corrisponde nel vettore personsId al valore - 1
	// perche' parte da 0
	public void addPerson(Person p) 
	{
		//set if the family has a MZ twin (children, ancestors are not considered)
		if (p.isMZ() && !p.isFounder())
			hasMZTwin = true;
			
		personsId.add(new Integer(p.id()));
		
		int index_to_be = persons.size(); //place it's occuping in the person list
		originalIndex.add(new Integer(index_to_be)); 
		p.setPersonIndex(index_to_be); 
		p.setOriginalIndex(index_to_be); 
		
		persons.add(p);	
	}
	

	public void initializeKinMatrix()
	{
		kinM = new double[persons.size()][];
		for (int x = 0; x < kinM.length; x++)
			kinM[x] = new double[(persons.size()) - x];
	}

	// Sort delle persone per avere i genitori sempre prima dei figli: serve per
	// il calcolo kinship da Lange
	public void sort() throws Exception
	{

		//Every family is composed by at least three subjects: child and two fakes parents
		Person tmpPerson, tmpFather, tmpMother;
		int fatherIndex, motherIndex, personIndex;
		
		// 1 - tutti gli ancestor hanno index = 0, set other index to 1
		for (Enumeration<Person> elements = persons.elements(); elements.hasMoreElements();)
		{
			tmpPerson = elements.nextElement();
			if (!tmpPerson.isFounder())
				tmpPerson.setIndex(1);
		}
		
		// 2 - tutti i figli dopo i genitori
		int maxInd = 1;
		boolean modified = true;
		while (modified) 
		{
			modified = false;
			for (int x = 0; x < persons.size(); x++) {
				tmpPerson =  persons.elementAt(x);
				if (!tmpPerson.isFounder()) 
				{
					personIndex = tmpPerson.index();
					tmpFather = persons.elementAt(tmpPerson.fatherIndex() - 1);
					tmpMother = persons.elementAt(tmpPerson.motherIndex() - 1);
					fatherIndex = tmpFather.index();
					motherIndex = tmpMother.index();
					if (!(personIndex > fatherIndex)) 
					{
						tmpPerson.setIndex(++personIndex);
						if (personIndex > maxInd)
							maxInd = personIndex;
						modified = true;
					}
					if (!(personIndex > motherIndex)) 
					{
						tmpPerson.setIndex(++personIndex);
						if (personIndex > maxInd)
							maxInd = personIndex;
						modified = true;
					}
				}
			}
		}
		
		// riindicizzo con i nuovi indici
		Collections.sort(persons, new PersonComparator());
		personsId.clear();
		originalIndex.clear();
		int newIndex = 1;
		for (int x = 0; x < persons.size(); x++) 
		{
			tmpPerson = persons.elementAt(x);
			tmpPerson.setPersonIndex(newIndex);
			newIndex++;

			personsId.add(new Integer(tmpPerson.id()));
			originalIndex.add(new Integer(tmpPerson.originalIndex()));
		}
		
		reindex(true);
	}

	// ricontrolla le persone, assegnando ad ognuna il padre e la madre
	// prelevati da idPersons
	public void reindex(boolean addChild) throws Exception
	{
		Person tmpPerson, tmpFather, tmpMother;
		int fatherPos, motherPos;
		
		for (int x = 0; x < persons.size(); x++) 
		{
			tmpPerson = persons.elementAt(x);

			if (!tmpPerson.isFounder()) //child
			{
				fatherPos = personsId.indexOf(new Integer(tmpPerson.father())) ;
				if (fatherPos < 0)
					throw new Exception("Subject " + tmpPerson.id() + " has no father.");
				tmpPerson.setFatherIndex(fatherPos + 1);
				
				motherPos = personsId.indexOf(new Integer(tmpPerson.mother())) ;
				if (motherPos < 0)
					throw new Exception("Subject " + tmpPerson.id() + " has no mother.");
				tmpPerson.setMotherIndex(motherPos + 1);
				
				if (addChild) 
				{
					tmpFather = persons.elementAt(fatherPos);
					tmpMother = persons.elementAt(motherPos);
				}
			}
		}
	}

	//return true if at least one family member is analyzable
	public boolean isAnalyzable(int trt, int snp) 
	{
		boolean result = false;
		for (Enumeration<Person> elements = persons.elements(); elements.hasMoreElements() && !result;)
		{
			Person p = elements.nextElement();	
			result = result || p.getAnalysis(trt, snp);
		}
		return result;
	}


	// calcola la matrice di kinship
	public void kinship() 
	{
		int num_persons = persons.size();
		
		//tmp matrix
		double[][] kinMatrix = new double[num_persons][];
		for (int i = 0; i < num_persons; i++)
			kinMatrix[i] = new double[num_persons - i];

		Person tmpPerson, tmpPerson2;
		int fatherPos, motherPos;
		double value;
		
		for (int x = 0; x < num_persons; x++) 
		{
			tmpPerson = persons.elementAt(x);
			if (tmpPerson.isFounder()) 
			{
				value = 0.5;
				kinMatrix = Utilities.setTriangMatrix(kinMatrix, value, x, x);
			} 
			else 
			{
				fatherPos = tmpPerson.fatherIndex() - 1;
				motherPos = tmpPerson.motherIndex() - 1;
				value = 0.5 + (0.5 *  Utilities.retrieveDTriangMatrix(kinMatrix, fatherPos, motherPos));
				kinMatrix = Utilities.setTriangMatrix(kinMatrix, value, x, x);

				for (int y = 0; y < x; y++) 
				{
					value = (0.5 * ( Utilities.retrieveDTriangMatrix(kinMatrix, fatherPos, y) + Utilities.retrieveDTriangMatrix(kinMatrix, motherPos, y)));
					kinMatrix = Utilities.setTriangMatrix(kinMatrix, value, x, y);
				}
			}
		}
		


		// per la variance component
		//FIXME Is it correct than the index i never sum up if affection is == 0?
		//Does the same hold for j?
		//It does not affect kinM, btw
		int i, j;
		i = 0;
		for (int x = 0; x < num_persons; x++) {
			tmpPerson = persons.elementAt(x);
			if (tmpPerson.affection() != 0) {
				j = i;
				for (int y = x; y < num_persons; y++) {
					tmpPerson2 = persons.elementAt(y);
					if (tmpPerson2.affection() != 0) {
						kinM = Utilities.setTriangMatrix(kinM,  Utilities.retrieveDTriangMatrix(kinMatrix, x, y), i, j);
						j++;
					}
				}
				i++;
			}
		}
		
	}
	
		
	public void mzTwinAdjust()
	{	
		double value = 0.5;
	
		int num_persons = persons.size();
		for (int i=0; i < num_persons; i++) 
		{
			Person p1 = persons.get(i);
			if (!p1.isMZ())
				continue;
					
			for (int j=1; j<num_persons; j++)
			{
				Person p2 = persons.get(j);
				if (p2.isMZ())
					kinM = Utilities.setTriangMatrix(kinM, value, i, j);
			}
		}
	}
	
	public void sanityCheck()
	{
		Vector<Integer> fathers = new Vector<Integer>();
	    Vector<Integer> mothers = new Vector<Integer>();
		
		for (Enumeration<Person> elements = persons.elements(); elements.hasMoreElements();)
		{
			Person current = elements.nextElement();
	
			if (!current.isFounder())
			{
				Integer mother = new Integer(current.mother());
				if(!personsId.contains(mother))
					mothers.add(mother);
				
				Integer father = new Integer(current.father());
				if(!personsId.contains(father))
					fathers.add(father);
			}
		}
		
		addMockParent(fathers, 1);
		addMockParent(mothers, 2);
	}
	
	//if parents are not available for a family, they are added
	private void addMockParent(Vector<Integer> parents, int gender)
	{
		for (Enumeration<Integer> elements = parents.elements(); elements.hasMoreElements();)
		{
			 //FIXME: affection is always 2
			int pid = elements.nextElement().intValue(); 
			this.addPerson(new Person(this.id, pid, 0, 0, gender, 2, "0", new Vector<Double>() ));
		}
	}

}


class PersonComparator implements Comparator<Person> 
{
	public int compare(Person p1, Person p2) 
	{
		int i1 = p1.index();
		int i2 = p2.index();

		return Math.abs(i1) - Math.abs(i2);
	}
}
