/*
 * 	 Normal.java
 * 
 * 
 *   famCNV: Copy Number Variant association for quantitative traits in families
 *   Copyright (C) 2010 Dr. Mario Falchi                                   

 *
 *   famCNV is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.

 *   famCNV is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with famCNV.  If not, see <http://www.gnu.org/licenses/>.
 *
 *   For any bugs or problems found, please contact us at
 *   m.falchi@imperial.ac.uk  
 *	 hariklia.eleftherohorinou06@imperial.ac.uk
 *	 visconti@di.unito.it
 */


class Normal 
{

	private double[] means;
	private double[] variances;
	private double[][][] varComponents;
	private double[][] linearModel;
	private double[][] varMatrix;

	private double[] scores;	
	private double[] residuals;
	
	Cholesky cholesky;
	
	private double likelihood;
	private double constant;
	
	private int dim, linearMD, vcCount;
	private boolean includeLikelihoodConstant, init;

	public double likelihood()
	{
		return this.likelihood;
	}

	public double[] variances()
	{
		return this.variances;
	}
	
	public double[] means()
	{
		return this.means;
	}

	public double[] scores() 
	{
		return this.scores;
	}
	
	public double scores(int index)
	{
		return this.scores[index];
	}
	
	public void resizeLM(int n)
	{
		linearMD = n;
	}
	
	public void setScores(int index, double value)
	{
		this.scores[index] = value;
	}
	
	public void setScores(double[] values)
	{
		this.scores = values;
	}
		
	public double[] linearModel(int index)
	{
		return this.linearModel[index];
	}
	
	public double[] predictor(int index)
	{
		double[] predictor = new double[dim];
		for(int i=0; i<dim; i++)
			predictor[i] = linearModel[i][index];
		
		return predictor;
	}
	
	public void setPredictor(double[] predictor, int index)
	{
		for(int i=0; i<dim; i++)
			linearModel[i][index] = predictor[i];
	}
	
	public void setLinearModel(int i, int j, double value)
	{
		linearModel[i][j] = value;
	}
	
	public double[][] varComponents(int index)
	{
		return this.varComponents[index];
	}
	
	public double varComponents(int i, int j, int k)
	{
		return this.varComponents[i][j][k];
	}
	
	public void setVarComponents(int i, int j, int k, double value)
	{
		varComponents[i][j][k] = value;
	}
	
	public int dim()
	{
		return dim;
	}
	
	public Normal(int vc, int lmd)
	{
		cholesky = new Cholesky();
		vcCount = vc;
		linearMD = lmd;
	}

	public void prepare(int dim) 
	{		
		includeLikelihoodConstant = false;
		init = false; 
		
		this.dim = dim;
		initDataStructures();
	}
	
	private void initDataStructures()
	{	
		constant = Math.log(2 * Math.PI) * -0.5 * dim;
		
		scores = new double[dim];			//Utilities.set(dim, 0.0);
		varMatrix = new double[dim][dim]; 	//Utilities.set(dim, dim, 0.0);
		residuals = new double[dim]; 		//Utilities.set(dim, 0.0);
		means = new double[linearMD]; 		//Utilities.set(linearMD, 0.0);
		variances = new double[vcCount];	//Utilities.set(vcCount, 0.0);
		linearModel = Utilities.set(dim, linearMD, 1.0);
		
		varComponents = new double[vcCount][dim][dim];
		varComponents[0] = Utilities.set(dim, dim, 0.0);	
		for(int i=0; i<dim; i++)
			varComponents[0][i][i] = 1.0;
	}
	
	
	public void setParameters(double[] mu, double[] sigma)
	{
		means = mu;
		variances = sigma;
	} 

	public void enableConstant()
	{
		if (!includeLikelihoodConstant)
		{
			includeLikelihoodConstant = true;
			if (init) likelihood += constant;
		}
	}

	public void disableConstant()
	{
		if (includeLikelihoodConstant)
		{
			includeLikelihoodConstant = false;
			if (init) likelihood -= constant;
		}
	}

	private void calculateResiduals() 
	{
		residuals = Utilities.set(dim, 0.0);
			
		for (int i = 0; i < dim; i++) 
			for (int j = 0; j < linearMD; j++)
				residuals[i] += linearModel[i][j] * means[j];
		
		for (int i = 0; i < dim; i++)
			residuals[i] -= scores[i];
			
	}

	private void calculateCovariances() 
	{
		varMatrix = Utilities.set(dim, dim, 0.0);
		
		for (int j = 0; j < vcCount; j++) 
			for (int r = 0; r < dim; r++) 
				for (int c = r; c < dim; c++)
					varMatrix[r][c] += variances[j] * varComponents[j][r][c];
	}


	public double evaluate() throws Exception
	{
		
		likelihood = includeLikelihoodConstant ? constant : 0.0;
		
		calculateResiduals();
		calculateCovariances();
		
		try 
		{
			cholesky.fastDecompose(varMatrix);
			cholesky.backSubst(residuals);
			
			likelihood -= 0.5 * Utilities.innerProduct(residuals, cholesky.x());
			likelihood -= cholesky.lnDeterminantL();
		}
		catch(Exception e)
		{
			throw e;
		}
		
		init = true;
		return likelihood;
	}

	
	public void DeleteColumn(int c) 
	{
		int nr = linearModel.length;
		int nc = linearModel[0].length - 1;
		double[][] tmpLinearModel = new double[nr][nc];
		for (int i = 0; i < nr; i++) 
		{
			int count = 0;
			for (int j = 0; j < nc; j++) 
				if (j != c)
					tmpLinearModel[i][count++] = linearModel[i][j];
		}
		
		linearModel = tmpLinearModel;
		means = new double[nc];
	}
	
	public double SSxy() throws Exception
	{
		return Utilities.innerProduct(cholesky.x(), residuals);
	}
	
	public double SSxx() throws Exception
	{
		double[] lrr = new double[linearModel.length];
		for (int i=0; i<lrr.length; i++)
			lrr[i] = linearModel[i][1];
		
		return Utilities.innerProduct(cholesky.x(), lrr);
	}
	
}
