/*
 * 	 NormalSet.java
 * 
 * 
 *   famCNV: Copy Number Variant association for quantitative traits in families
 *   Copyright (C) 2010 Dr. Mario Falchi                                   

 *
 *   famCNV is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.

 *   famCNV is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with famCNV.  If not, see <http://www.gnu.org/licenses/>.
 *
 *   For any bugs or problems found, please contact us at
 *   m.falchi@imperial.ac.uk  
 *	 hariklia.eleftherohorinou06@imperial.ac.uk
 *	 visconti@di.unito.it
 */

class NormalSet 
{

	private Normal[] sets;
	private double[] variances;
	private double[] means;
	
	private int size;
	private int linearMD;
	private int vcCount;

	private double precision;
	private double likelihood;

	public NormalSet(int setCount, int lmd, int vc) 
	{
		likelihood = 0.0;
		precision = Utilities.PRECISION;
		linearMD = lmd;
		vcCount = vc;
		size = setCount;
		
		init();
	}
	
	private void init()
	{
		means = new double[linearMD]; //Utilities.set(linearMD, 0.0);
		
		sets = new Normal[size];
		for (int i = 0; i < size; i++)
			sets[i] = new Normal(vcCount, linearMD);
	}
	
	public Normal sets(int index) 
	{
		return sets[index];
	}
	
	public double[] variances()
	{
		 return variances;
	}
	
	public double[] means()
	{
		 return means;
	}
	
	public double means(int i)
	{
		 return means[i];
	}
	
	public double variances(int i)
	{
		 return variances[i];
	}

	public void solve() throws Exception 
	{
		
		try
		{
			editLinearDegenerates();
		}
		catch(Exception e)
		{
			throw e;
		}
			
		int parameters = countParameters();

		Amoeba solver = new Amoeba(this, parameters);
		solver.reset(parameters, 1);
		solver.setPoint(getStartingPoint());  
		
		try 
		{
			solver.minimize(precision); 
			
			double lastmin;
			double scale = 2.0;
			int counter = 0;
			
			do 
			{		
				double varMax = solver.point(linearMD);
	
				for (int i = linearMD + vcCount - 1; i > linearMD; i--) 
					if (solver.point(i) > varMax)
						varMax = solver.point(i);
				
				for (int i = linearMD + vcCount - 1; i >= linearMD; i--) 
					if (solver.point(i) < (varMax - 5.0))
						solver.setPoint(i, (varMax - 5.0));
				
				lastmin = solver.fmin();
				
				scale *= 0.5;
				
				solver.reset(parameters, scale);
				
				solver.minimize(precision);
				counter++;
			} while (solver.fmin() > precision && ((lastmin - solver.fmin()) / solver.fmin()) > precision);

			selectPoint(solver.point());
			
		} 
		catch (Exception e) 
		{
			throw e;
		}
		
		
	}

	private void editLinearDegenerates() throws Exception
	{
		int rows = countObservations();
		int cols = linearMD;
		double[][] m = new double[rows][cols]; 	//Utilities.set(rows, cols, 0.0); 
		double[] b = new double[rows]; 			//Utilities.set(rows, 0.0);
		
		int total = 0;
		for (int i = 0; i < size; i++) 
		{ 
			for (int j = 0; j < sets[i].dim(); j++) 
			{ 
				b[total] = sets[i].scores(j);
				m[total] = sets[i].linearModel(j);
				total++;
			}
		}
		
		SVD engine = new SVD();
		
		int c = 1;
		while (c <= cols) 
		{	
			try
			{
				engine.decompose(m, rows, c);
			}
			catch(Exception e)
			{
				throw e;
			}
			
			engine.edit();
			boolean redundant = false;
			
			for (int j = 0; j < c; j++) 
				if (engine.w(j) == 0.0)
					redundant = true;
			
			if (redundant) 
			{
				for (int i = 0; i < size; i++)
					sets[i].DeleteColumn(c-1);

				double[][] tmpM = new double[rows][cols - 1];
				
				for (int i = 0; i < rows; i++) 
				{
					int count = 0;
					for (int j = 0; j < cols; j++) 
					{
						if (j != c - 1)
							tmpM[i][count++] = m[i][j];
					}
				}
				
				m = tmpM;
				cols--; 
				
				if (c > cols)
				{
					try
					{
						engine.decompose(m, rows, cols);
					}
					catch(Exception e)
					{
						throw e;
					}
				}
			} 
			else
				c++;
		}
					
		variances = new double[vcCount];
		engine.backSubst(b);
			
		int tmp = means.length; //engine may change the lenght of the means vector
		means = engine.x(); 
		if (means.length != tmp)
		{
			tmp = means.length;
			linearMD = tmp; 
			for (int i = 0; i < size; i++)
				sets[i].resizeLM(tmp);
		}

		double residualVar = (Math.abs(engine.RSS(m, b)) + Utilities.TINY) / ((double) rows);
		double[] var_weights = Utilities.set(vcCount, 0.0); 
		int[] var_counts = Utilities.set(vcCount, 0); 
		double[] scaled_variances = new double[vcCount];

		for (int i = 0; i < size; i++) 
		{
			for (int j = 0; j < sets[i].dim(); j++) 
			{
				double scaled_variances_sum = 0.0;
				for (int v = 0; v < vcCount; v++) 
				{
					if (sets[i].varComponents(v).length > 0)
						scaled_variances[v] = sets[i].varComponents(v,j,j); 
					else
						scaled_variances[v] = 0.0;
					
					scaled_variances_sum += scaled_variances[v];
				}
				
				double value = 1.0 / scaled_variances_sum;
				for (int v = 0; v < vcCount; v++)
					scaled_variances[v] *= value;
				
				for (int v = 0; v < vcCount; v++) {
					if (scaled_variances[v] > 0.0) 
					{
						var_weights[v] += scaled_variances[v];
						var_counts[v]++;
					}
				}
			}	
		}
		
		for (int v = 0; v < vcCount; v++) 
			variances[v] = residualVar * var_weights[v] / var_counts[v];

	}


	private double[] getStartingPoint() 
	{
		
		double[] startPoint = new double[linearMD + vcCount];
		
		for(int i=0; i<linearMD; i++)
			startPoint[i] = means[i];
			
		for(int i=0; i<vcCount; i++)
			startPoint[i+linearMD] = Math.log(variances[i]);

		return startPoint;	
	}
	
	
	private void selectPoint(double[] v) 
	{	
		int j = linearMD;
		for(int i=0; i<j; i++)
			means[i] = v[i];
		
		for(int i=0; i<vcCount; i++)
		{
			if(v[j] > 16 || v[j] < -16)
				variances[i] = v[j] > 0 ? 1.0E7 : 1.0E-7;
			else
				variances[i] = Math.exp(v[j]);
			j++;
		}
		
		for(int i=0; i<size; i++)
			sets[i].setParameters(means, variances);
	}


	public int countParameters() 
	{
		return linearMD + vcCount;
	}

	private int countObservations() 
	{
		int rows = 0;
		for (int i = 0; i < size; i++)
			rows += sets[i].dim();
		
		return rows;
	}

	public double callEvaluate(double[] point) throws Exception
	{
		try 
		{
			selectPoint(point);
			return evaluate();
		}
		catch (Exception e)
		{
			throw e;
		}
	}

	public void enableConstant() 
	{
		for (int i = 0; i < size; i++)
			sets[i].enableConstant();
	}
	
	public void disableConstant() 
	{
		for (int i = 0; i < size; i++)
			sets[i].disableConstant();
	}
	
	public double evaluate() throws Exception 
	{
		try
		{
			likelihood = 0.0;
			for (int i = 0; i < size; i++)
				likelihood += sets[i].evaluate(); //operator is NORMAL_MUL_LK
			
			return (-likelihood);
		}
		catch (Exception e)
		{
			throw e;
		}
	}
	
	
	public double[] evaluateDetails() 
	{
		double[] LK = new double[size];

		for (int i = 0; i < size; i++)
			LK[i] = (-sets[i].likelihood());
		
		return LK;
	}


	public double SSxy() throws Exception
	{
		double ssxy = 0.0;
		for (int i = 0; i < size; i++)
			ssxy += sets[i].SSxy();
			
		return ssxy;
	}
	
	public double SSxx() throws Exception
	{
		double ssxx = 0.0;
		for (int i = 0; i < size; i++)
			ssxx += sets[i].SSxx();
			
		return ssxx;
	}

}

