/*
 * 	 PedParser.java: Class to read the input file *.ped
 * 
 * 
 *   famCNV: Copy Number Variant association for quantitative traits in families
 *   Copyright (C) 2010 Dr. Mario Falchi                                   

 *
 *   famCNV is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.

 *   famCNV is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with famCNV.  If not, see <http://www.gnu.org/licenses/>.
 *
 *   For any bugs or problems found, please contact us at
 *   m.falchi@imperial.ac.uk  
 *	 hariklia.eleftherohorinou06@imperial.ac.uk
 * 	 visconti@di.unito.it
 */



import java.io.*;
import java.util.*;

class PedParser 
{
	private int counter;
	private String[] pedData;
	private String[] phenoData;
	
	static int INFO_SIZE = 5;

	public PedParser(String fileName, String phenofileName)  throws Exception
	{
		try 
		{
			pedData = readFile(fileName);
		}
		catch(IOException e)
        {
			throw new Exception("ERROR: Stopping as input file was not found ");
		}
			
		if (Constants.pheno != null) 
		{
			try 
			{
				phenoData = readFile(phenofileName);
			}
			catch(Exception e)
			{
				throw new Exception("ERROR: Check --pheno argument or phenotypes file");
			}
		}
		
		counter = 0;
	}
	
	public boolean endOfData()
	{
		return counter >= pedData.length;
	}

	private static String[] readFile(String filename)  throws IOException
	{
		FileArrayProvider fap = new FileArrayProvider();
        String[] lines = fap.readLines(filename);
        return lines;
	}


	public Person readIndividual() throws Exception
	{
		if (endOfData())
			return null;
			
		String s = pedData[counter];
		
		int famId, id, fatherId, motherId, sex, affection;
		String twin;
		Vector<Double> traits = new Vector<Double>();

		try 
		{
			StringTokenizer t = new StringTokenizer(s);
			if (t == null)
				t = new StringTokenizer(s, "\t");
				
			int[] info = new int[INFO_SIZE];
			int c = 0;
			while (t.hasMoreTokens() && c < INFO_SIZE)
			{
				info[c] = Integer.parseInt(t.nextToken());
				c++;
			}
			
			//the line is not valid: some information is missing
			if (c < INFO_SIZE) //affections and twins aren't int!
			{
				throw new Exception("ERROR: Not well-formed PED file.\n\tCheck PED file at line " + (counter+1));
			}
				
			famId = info[0]; // column1
			id = info[1]; // column2
			fatherId = info[2]; // column3
			motherId = info[3]; // column4	
			sex =  info[4]; // column5
			
			
			if (!Utilities.isValid(new Integer(sex), Utilities.VALID_SEX))
				throw new Exception("ERROR: Sex information is not valid.\n\tCheck PED file at line " + (counter+1));
			
			if (t.hasMoreTokens())
				t.nextToken(); //phenotype token, that is column6, is not considered
			else throw new Exception("ERROR: phenotype information is not present.\n\tCheck PED file at line " + (counter+1));
			
			if (t.hasMoreTokens())
				twin = t.nextToken(); // column7
			else throw new Exception("ERROR: Twin information is not present.\n\tCheck PED file at line " + (counter+1));
			
			if (!Utilities.isValid(twin, Utilities.VALID_TWIN))
				throw new Exception("ERROR: Twin information is not valid.\n\tCheck PED file at line " + (counter+1));
			
			affection = 2; //FIXME (see also Family:345)
			
			// IF PHENOTYPES FILE IS SPECIFIED DO NOT USE PED FILE
			if (Constants.pheno != null)
			{
				try 
				{
					affection = readPheno(id);
				}
				catch(Exception e)
				{
					throw e;
				}
				
				for (int i = 0; i < Constants.mpheno; i++)
					t.nextToken();
			}

			//Continue t parsing,otherwise
			while (t.hasMoreTokens()) 
			{
				String trait = new String(t.nextToken());
				try 
				{
					if (trait.equals("X") || trait.equals("NA") || trait.equals("x")) 
					{
						// if trait is missing set affection to missing
						traits.add(new Double(Utilities.INVALID_D));// THINK IT'S WRONG!!! FIXME
						affection = Utilities.INVALID_I;
					} 
					else
						traits.add(new Double(trait));
				} 
				catch (Exception exc) 
				{
					throw new Exception("ERROR: Check traits.\n\tWrong trait code for individual " + id);
				}
			}

			counter++; 
			return new Person(famId, id, fatherId, motherId, sex, affection, twin, traits);
		}
		catch (NumberFormatException e)
		{
			throw new Exception("ERROR: Check PED file at line " + (counter+1));
		}
		catch (Exception exception) {
			throw exception;
		}
		
		
	}
	
	
	private int readPheno(int id) throws Exception
	{
		String phs = phenoData[counter];
		StringTokenizer pht = new StringTokenizer(phs);
		int affection = 2;	//FIXME
		
		//skip family and individual
		int i = 0;
		while (pht.hasMoreTokens() && i < 2) 
		{
			i++;
			pht.nextToken()	;
		}
		
		//the line is not valid: some information is missing
		if (i < 2)
		{
			throw new Exception("ERROR: Check phenotypes file.");
		}
		
		if (pht.hasMoreTokens()) 
		{
			for (int loop = 0; loop < Constants.mpheno; loop++) {
				//FIXME: is correct that the affection ia alway set to 2?
				//is it correct that, when mpheno is > 1 the first/last token is return?
				String aff = "2"; //new String(pht.nextToken());
				if (aff.equalsIgnoreCase("X")
						|| aff.equalsIgnoreCase("NA")
						|| aff.equalsIgnoreCase("0")
						|| aff.equalsIgnoreCase("-9"))
					affection = Utilities.INVALID_I;
				else if (aff.equalsIgnoreCase("1")
						|| aff.equalsIgnoreCase("2"))
					affection = Integer.parseInt(aff);
				else {
					throw new Exception("ERROR: Check phenotypes column.\n\tWrong phenotype code for individual " + id);
				}
				//FIXME: which return?
			}
		}
		else {
			throw new Exception("ERROR: Check phenotypes file.");
		}
		
		return affection;
		
	}
}
