/*
 * 	 QuantitativeData.java
 * 
 * 
 *   famCNV: Copy Number Variant association for quantitative traits in families
 *   Copyright (C) 2010 Dr. Mario Falchi                                   

 *
 *   famCNV is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.

 *   famCNV is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with famCNV.  If not, see <http://www.gnu.org/licenses/>.
 *
 *   For any bugs or problems found, please contact us at
 *   m.falchi@imperial.ac.uk  
 *   visconti@di.unito.it
 */

import java.util.*;
import java.lang.Runtime;
import java.util.concurrent.*;


public class QuantitativeData 
{
    private Vector<Person>  pedigree;
    private Vector<Integer> individuals;
    private int nTraits;
    
    public QuantitativeData(Vector<Person> ped, int n) 
    {
	pedigree = ped;
	nTraits = n;
	
	fillIndividual();
    }


    private void fillIndividual()
    {
	individuals = new Vector<Integer>();
	
	for (int t = 0; t < nTraits; t++)
	    for (int i = 0; i < pedigree.size(); i++)
		if (pedigree.get(i).traits().get(t).doubleValue() != Utilities.INVALID_D)
		    individuals.add(new Integer(i));
    }
    


    public void inverseNormal() throws Exception
    {
	Vector<Integer> inputs = new Vector<Integer>();
	
	for (int t=0; t<nTraits; t++)
	    inputs.add(new Integer(t));

	processInputs(inputs);
    }


    public void processInputs(Vector<Integer> inputs) throws InterruptedException, ExecutionException, Exception
    {
	
	int threads = Runtime.getRuntime().availableProcessors();
	ExecutorService service = Executors.newFixedThreadPool(threads);

	for (final Integer input : inputs)
	{
	    Callable<Object> callable = new Callable<Object>()
	    {
		public Object call() throws Exception
		{
		    double scale = 1.0 / individuals.size();
		    double[] zs = new double[individuals.size()];
		    
		    int t = input.intValue();
		    Double[] v = new Double[individuals.size()];	    
		    for (int i = 0; i < individuals.size(); i++)
			v[i] = pedigree.get(individuals.get(i).intValue()).traits().get(t);
		    
		    RankVector rankVector = new RankVector(v);
		    
		    int j = -1;
		    for(int i=0; i < individuals.size(); i++)
		    {
			for(j = i; j < individuals.size() - 1; j++)
			    if (rankVector.get(rankVector.getRank(i)) != rankVector.get(rankVector.getRank(j)))
				break;
	
			if (rankVector.get(rankVector.getRank(i)) != rankVector.get(rankVector.getRank(j)))
				j--;
	
			double z = Utilities.ninv(((i + j) * 0.5 + 0.5) * scale);
			
			for (int k = i; k <= j; k++)
			    zs[k] = z;
	    
			i=j;
	
		    }
	
		    for(int i=0; i<individuals.size(); i++)
		    {
			 int index = individuals.get(i).intValue();
			 Person p = pedigree.get(index);
			 p.traits().set(t, zs[rankVector.getRank(i).intValue()]);
		    }

		    return null;
		};

	    };

	    service.submit(callable);
	}

	service.shutdown();
    }

}


class RankVector
{
    private Double[] vector;
    private Integer[] rank;


    RankVector(Double[] v)
    {
	vector = v;
	rank = new Integer[v.length];
	doRank();
    }


    public Double get(int i)
    {
	return vector[i];
    }

    public Integer getRank(int i)
    {
	return rank[i];
    }


    
    private void doRank()
    {
	

	ArrayComparator<Double> comparator = new ArrayComparator<Double>(vector);
	rank = comparator.createIndexArray();
    }

}



class ArrayComparator<T extends Comparable<? super T>>
{
    private final T[] array;
    private final SortedMap<T, Vector<Integer>> sortedArray;

    public ArrayComparator(T[] array)
    {
        this.array = array;
        sortedArray = new TreeMap<T, Vector<Integer>>();
        for(int i = 0 ; i < array.length ; i ++)
	{
	    Vector<Integer> tmp = sortedArray.get(array[i]);
	    if (tmp == null)
		tmp = new Vector<Integer>();

	    tmp.add(Integer.valueOf(i));
	    sortedArray.put(array[i], tmp);
	}
        
    }

    public Integer[] createIndexArray()
    {
        Integer[] indexArray = new Integer[array.length];
        int i = 0;

        for(T key : sortedArray.keySet())
	{

	    Vector<Integer> tmp = sortedArray.get(key);
	    for (int j=0; j<tmp.size(); j++)
	    {
		int index = tmp.get(j).intValue();
		indexArray[index] = i;
		i++;
	    }
	    
        }
	
        return indexArray;
    }
}

