/*
 * 	 vc.java
 * 
 * 
 *   famCNV: Copy Number Variant association for quantitative traits in families
 *   Copyright (C) 2010 Dr. Mario Falchi                                   

 *
 *   famCNV is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.

 *   famCNV is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with famCNV.  If not, see <http://www.gnu.org/licenses/>.
 *
 *   For any bugs or problems found, please contact us at
 *   m.falchi@imperial.ac.uk  
 *	 hariklia.eleftherohorinou06@imperial.ac.uk
 *   visconti@di.unito.it
 */

import mathutils.ProbDist;
import java.util.*;

class VC 
{
	private int numInd;
	private int index;
	
	private int trt;
	private int snp;

	private NormalSet normalNull, normalFull;
	private int linearNull, linearFull;   
	private int numCovar;
	private Vector<Double> lrr; //used for store value for var(lrr) evaluation
	private Vector<Double> ys; //used for store value for var(Y) evaluation
	
	private double posF;
	private double giniC;
	    

	public VC(int numFam, int numCovar, int t, int s) 
	{
		trt = t;
		snp = s;
		
		this.numCovar = numCovar;
		setLinearParams();
		
		normalNull = new NormalSet(numFam, linearNull, Utilities.VC_NULL);
		normalFull = new NormalSet(numFam, linearFull, Utilities.VC_FULL);
		
		index = -1;
		numInd = 0;	
		
		lrr = new Vector<Double>();
		ys = new Vector<Double>();
		
		posF = Utilities.INVALID_D;
		giniC = Utilities.INVALID_D;
	}
	
	
	private void setLinearParams()
	{	
		linearFull = numCovar + 2;
		linearNull = numCovar + 1;  
		
		if (Constants.Ball) 
		{
			linearFull++;
			linearNull++;
		} 	
	}


	public void fill(Family f)
	{
		prepare(f);
		
		if (numCovar > 0)
			fillCovariates(f);
		
		fillKIN3(f);
		fillPhenotIBS(f);
		fillLogRBAll(f);
	}


	public void prepare(Family fam) 
	{
		int numIndFam = 0;
		index++;
			
		for (Enumeration<Person> elements = fam.getPersons().elements(); elements.hasMoreElements();)
		{
			Person tmpPerson = elements.nextElement();	
			if (tmpPerson.getAnalysis(trt, snp))
				numIndFam++;
		}

		(normalFull.sets(index)).prepare(numIndFam);
		(normalNull.sets(index)).prepare(numIndFam);

		numInd += numIndFam;
	}


	public void fillKIN3(Family fam) 
	{
		Person person;
		
		double[][] kinValues = fam.getKinM();
		
		int ik = 0;		
		for (int i = 0; i < fam.getNumIndividuals(); i++) 
		{
			person = fam.getIndividual(i);
			if (person.getAnalysis(trt, snp)) 
			{
				int jk = ik;
				for (int j = i; j < fam.getNumIndividuals(); j++) 
				{
					person = fam.getIndividual(j);
					if (person.getAnalysis(trt, snp)) 
					{	
						double value = 2 *  Utilities.retrieveDTriangMatrix(kinValues, i, j);

						(normalNull.sets(index)).setVarComponents(1, ik, jk, value);
						(normalNull.sets(index)).setVarComponents(1, jk, ik, value);
						
						(normalFull.sets(index)).setVarComponents(1, ik, jk, value);
						(normalFull.sets(index)).setVarComponents(1, jk, ik, value);
						
						jk++;
					}
				}
				ik++;				
			}
		}
	}


	public void fillPhenotIBS(Family fam) 
	{
		int l = 0;
		for (int x = 0; x < fam.getNumIndividuals(); x++) 
		{
			Person tmpPerson = fam.getPersons().elementAt(x);
			if (tmpPerson.getAnalysis(trt, snp)) 
			{
				Double trait = tmpPerson.traits().get(trt);
				double val = trait.doubleValue();
				
				ys.add(trait);
				
				(normalNull.sets(index)).setScores(l, val);
				(normalFull.sets(index)).setScores(l, val);
				l++;
			}
		}
	}

	public void fillCovariates(Family fam) 
	{
		int l = 0;
		for (int x = 0; x < fam.getNumIndividuals(); x++) 
		{
			Person p = fam.getPersons().elementAt(x);
			if (p.getAnalysis(trt, snp)) 
			{
				Vector<Double> covariates = p.covariates();
				for (int c = 0; c < covariates.size(); c++) 
				{
					double cov = (covariates.elementAt(c)).doubleValue();
					(normalFull.sets(index)).setLinearModel(l, (linearFull-numCovar+c), cov);
					(normalNull.sets(index)).setLinearModel(l, (linearNull-numCovar+c), cov);
				}
				l++;
			}
		}
	}

	public void fillLogRBAll(Family fam) 
	{
		int l = 0;
		for (int x = 0; x < fam.getNumIndividuals(); x++) 
		{
			Person tmpPerson = fam.getPersons().elementAt(x);
			if (tmpPerson.getAnalysis(trt, snp)) 
			{
				lrr.add(new Double(tmpPerson.logR(snp)));
				
				if (Constants.Ball) 
				{
					(normalNull.sets(index)).setLinearModel(l, (linearNull-numCovar-1), tmpPerson.bAllele(snp));
					
					(normalFull.sets(index)).setLinearModel(l, (linearFull-numCovar-1), tmpPerson.bAllele(snp));
					(normalFull.sets(index)).setLinearModel(l, (linearFull-numCovar-2), tmpPerson.logR(snp));
				} 
				else
					(normalFull.sets(index)).setLinearModel(l, (linearFull-numCovar-1), tmpPerson.logR(snp));

				l++;
			}
		}

	}


	public Result solve() throws Exception 
	{			
		try 
		{
			normalNull.disableConstant();
			normalFull.disableConstant();
			
			normalNull.solve();
			normalFull.solve();
		
			normalNull.enableConstant();
			double statisticNull = normalNull.evaluate();
			int dfNull = numInd - normalNull.countParameters();

			normalFull.enableConstant();
			double statisticFull = normalFull.evaluate();
			int dfFull = numInd - normalFull.countParameters();
	
			double testValue = 2 * (statisticNull - statisticFull);
			testValue = (testValue < 0) ? 0 : testValue;

			double pvalue = ProbDist.gammq(0.5 * 1, 0.5 * testValue);
			if (Constants.relc != Utilities.INVALID_D && pvalue <= Constants.relc) 
				LKstats(normalFull.evaluateDetails(), normalNull.evaluateDetails());
		
			//variance explained and beta (variance of lrr is the genotype variance)	
	        double beta = normalFull.means(1);
	        double ve = beta * beta * Utilities.variance(lrr, true) / Utilities.variance(ys, true);
			
			double epvalue = 1;
			for (int permutation=0; permutation<Constants.permutations; permutation++)
			{
				double epv = permute(statisticNull);
				if (epv <= pvalue)
					epvalue++;
			}
			epvalue = (Constants.permutations == -1) ? Utilities.INVALID_D : epvalue/(Constants.permutations + 1);
		
			return new Result(statisticNull, statisticFull, dfNull, dfFull, testValue, pvalue, epvalue, posF, giniC, beta, ve, normalNull.variances(), normalFull.variances());
			
		}
		catch(Exception e)
		{
			return new Result(e.getMessage());
		}
	}
	

	private double permute(double statisticNull)
	{
		try
		{
			for (int i=0; i<index; i++)
			{
				//FIXME: should BAF be shuffled as well?		
				double[] slrr = Utilities.shuffle((normalFull.sets(index)).predictor(1)); //LRR			
				(normalFull.sets(index)).setPredictor(slrr, 1);
			}
			
			normalFull.disableConstant();
			normalFull.solve();
			normalFull.enableConstant();
			double statisticFull = normalFull.evaluate();
			
			double testValue = 2 * (statisticNull - statisticFull);
			testValue = (testValue < 0) ? 0 : testValue;

			return ProbDist.gammq(0.5 * 1, 0.5 * testValue);
		}
		catch (Exception e)
		{
			//doing nothing, is the system can't be solved I assume
			//that the pvalue of the permutation is higher than that
			//of the real model
			return 1.1;
		}
	}

	private double varianceExplained() throws Exception
	{
		// Numerator of test statistic is Genotype * SIGMA^-1 * Phenotype (SSxy)
	    // Denominator of test statistic is Genotype * SIGMA^-1 * Genotype (SSxx)
        double assoc_effect = - normalFull.SSxy() / normalFull.SSxx();
		return assoc_effect * assoc_effect * Utilities.variance(lrr, true) / totalVariance();
	}
	
	
	private void LKstats(double[] famLKfull, double[] famLKnull) 
	{
		double sum = 0;
		double gsum = 0;
		int posit = 0;
		double[] valuesAll = new double[famLKfull.length];

		for (int i = 0; i < famLKfull.length; i++) 
		{
			valuesAll[i] = famLKnull[i] - famLKfull[i];
			if (valuesAll[i] > 0)
				posit++;
		}
		
		double[] values = new double[posit];
		
		int idx = 0;
		for (int i = 0; i < famLKfull.length; i++)
			if (valuesAll[i] > 0)
				values[idx++] = valuesAll[i];

		Arrays.sort(values);

		for (int i = 0; i < posit; i++) 
		{
			sum += values[i];
			gsum += 2 * values[i] * (i + 1);
		}

		sum = sum * values.length;

		giniC = (gsum/sum) - ((double)(posit+1)/posit);
		posF = (double) posit / valuesAll.length;
	}
	
	public double totalVariance()
	{
		double[] variances = normalFull.variances();
		double variance = 0.0;
		for(int i=0; i<variances.length; i++)
			variance += variances[i];
		
		return variance;
	}
	
	public double assoc_effect()
	{
		return normalFull.means(1);
	}


}
